package grizzly.tutorial.conncaching.client;

import static grizzly.tutorial.conncaching.Utils.*;

import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * This class opens a listener port and accepts responses from all servers
 * that have been sent requests earlier. Response is dispatched per server id.
 * 
 * Client code registers a per server handler in processing the response
 * message. The handler may emulate a client ORB behavior in generating
 * object-method invocations, but not a concern here.
 * 
 * Interesting article: http://forum.java.sun.com/thread.jspa?threadID=579675&messageID=2940535
 * if tested on windows 2000.
 * 
 * @author wenboz
 */
public class ResponseDispatchChannel {
	
	private InetSocketAddress clientId;

	private ServerSocketChannel ssc;

	private Selector selector;

	Map<String, ResponseHandler> handlers = new HashMap<String, ResponseHandler>();

	public ResponseDispatchChannel(InetSocketAddress clientId) {
		this.clientId = clientId;
	}

	public void open() throws Exception {

		ssc = ServerSocketChannel.open();
		ServerSocket serverSocket = ssc.socket();
		serverSocket.bind(clientId);
		selector = Selector.open();

		ssc.configureBlocking(false);
		ssc.register(selector, SelectionKey.OP_ACCEPT);
	}

	public void close() {

		if (selector != null) {
			try {
				selector.close();
			} catch (Exception ex) {
				trace(ex);
			} finally {
				selector = null;
			}
		}

		if (ssc != null) {
			try {
				ssc.close();
			} catch (Exception ex) {
				trace(ex);
			} finally {
				ssc = null;
			}
		}
	}

	/**
	 * This is an interruptable blocking call!
	 */
	public void dispatch() throws Exception {

		SelectionKey key;
		SocketChannel sc;
		while (true) {
			if (Thread.interrupted()) {
				throw new InterruptedException();
			}
			int n = selector.select();
			if (n == 0) {
				continue;
			}
			Iterator<SelectionKey> it = selector.selectedKeys().iterator();

			while (it.hasNext()) {
				key = it.next();
				if (key.isAcceptable()) {
					sc = ssc.accept();
					if (sc != null) {
						sc.configureBlocking(false);
						handleResponse(sc);
					}
				}

				it.remove();
			}
		}
	}

	/**
	 * Asynchronously handled.
	 * 
	 * todo: cleanup (or does client app need it)
	 */
	private void handleResponse(final SocketChannel sc) throws Exception {

		Thread handler = new Thread(new Runnable() {

			private ByteBuffer buffer = ByteBuffer.allocateDirect(1024);
			
			String msgBuffer = "";

			public void run() {

				Selector readSelector = null;
				try {
					readSelector = Selector.open();
					SelectionKey tmpKey = sc.register(readSelector,
							SelectionKey.OP_READ);
					int code;

					while (sc.isConnected()) {
						tmpKey.interestOps(tmpKey.interestOps()
								| SelectionKey.OP_READ);
						code = readSelector.select(); // blocking  
						tmpKey.interestOps(tmpKey.interestOps()
								& (~SelectionKey.OP_READ));
						if (code != 0) {
							readSelector.selectedKeys().clear();        
							buffer.clear();
							int count;
							while ((count = sc.read(buffer)) > 0 && buffer.hasRemaining()) {
							}

							buffer.flip();
							String msg = readBuffer(buffer);
							trace("handle message: " + msg);
							handleResponse(msg);

							if (count == -1) {
								break;  // eof
							}
						}
					}
				} catch (Exception ex) {
					trace(ex);
				} finally {
					try {
						sc.close();
					} catch (Exception exx) {
					}
					if (readSelector != null) {
						try {
							readSelector.close();
						} catch (Exception ex) {
						}
					}
				}
			}
			
			private void handleResponse(String msg) {
				
				try {
					msgBuffer = msgBuffer + msg;
					
					if (msgBuffer.indexOf(PAYLOAD_DELIMITER) != -1) {
						String[] messages = msgBuffer.split(PAYLOAD_DELIMITER);
						
						for (int i = 0; i < messages.length -1; i ++) {
							handleAtomicResponse(messages[i]);
						}
						
						String lastMessage = messages[messages.length -1];
						if (msgBuffer.endsWith(PAYLOAD_DELIMITER)) {
							handleAtomicResponse(lastMessage);
							msgBuffer = "";
						}
						else {
							msgBuffer = lastMessage; 
							
						}
					}
				}
				catch (Exception ex) {
					trace(ex);
				}
			}
			
			private void handleAtomicResponse(String atomicMessage) throws Exception {
				int pos = atomicMessage.indexOf("\n");
				String head = atomicMessage.substring(0, atomicMessage.indexOf("\n"));
				String content = atomicMessage.substring(pos+1);
				
				ResponseHandler handler = handlers.get(head);
				if (handler != null) {
					handler.handle(content);
				}
				
			}
		});

		handler.setDaemon(true);
		handler.start();
	}

	public InetSocketAddress getClientId() {
		return clientId;
	}

	public void register(InetSocketAddress serverId, ResponseHandler handler) {
		handlers.put(serverId.toString(), handler); 
	}

	public void unregister(InetSocketAddress serverId) {
		handlers.remove(serverId); 
	}
}
