package grizzly.tutorial.conncaching.server;

import static grizzly.tutorial.conncaching.Utils.PAYLOAD_DELIMITER;
import static grizzly.tutorial.conncaching.Utils.readBuffer;
import static grizzly.tutorial.conncaching.Utils.resolveAddress;
import static grizzly.tutorial.conncaching.Utils.resolveAddressParam;
import static grizzly.tutorial.conncaching.Utils.trace;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.AlreadyConnectedException;
import java.nio.channels.SelectionKey;
import java.util.HashMap;
import java.util.Map;

import com.sun.grizzly.CallbackHandler;
import com.sun.grizzly.ConnectorHandler;
import com.sun.grizzly.ConnectorHandlerPool;
import com.sun.grizzly.Context;
import com.sun.grizzly.Controller;
import com.sun.grizzly.DefaultProtocolChain;
import com.sun.grizzly.DefaultProtocolChainInstanceHandler;
import com.sun.grizzly.IOEvent;
import com.sun.grizzly.ProtocolChain;
import com.sun.grizzly.ProtocolFilter;
import com.sun.grizzly.SelectionKeyHandler;
import com.sun.grizzly.TCPSelectorHandler;
import com.sun.grizzly.connectioncache.client.CacheableConnectorHandlerPool;
import com.sun.grizzly.connectioncache.server.CacheableSelectionKeyHandler;
import com.sun.grizzly.filter.ReadFilter;
import com.sun.grizzly.util.WorkerThread;

/**
 * The server listens to a single port and dispatch the request according to the
 * object id (hence server being a naming too) and the message (params). And for
 * simiplicity we leave out the remote method invocation semantics at the MORB
 * level. The server object handler can handle the message directly.
 * 
 * todo: error-proof shutdown cleanup.
 * 
 * @author wenboz
 */
public class ServerRequestController {

	private String serverId;

	private InetSocketAddress localAddr;

	Map<String, RequestHandler> handlers = new HashMap<String, RequestHandler>();

	private Controller grizzlyController;

	public ServerRequestController(String serverId) {
		this.serverId = serverId;
	}

	public void open(String addr) throws Exception {

		this.localAddr = resolveAddressParam(addr);

		grizzlyController = new Controller();

		ConnectorHandlerPool cacheableHandlerPool = new CacheableConnectorHandlerPool(
				grizzlyController, 5, 2, 2);
		grizzlyController.setConnectorHandlerPool(cacheableHandlerPool);

		SelectionKeyHandler cacheableKeyHandler = new CacheableSelectionKeyHandler(
				5, 2);
		grizzlyController.setSelectionKeyHandler(cacheableKeyHandler);

		TCPSelectorHandler tcpHandler = new TCPSelectorHandler();
		final MessageDecoder filter = new MessageDecoder();

		tcpHandler.setInet(localAddr.getAddress());
		tcpHandler.setPort(localAddr.getPort());

		grizzlyController
				.setProtocolChainInstanceHandler(new DefaultProtocolChainInstanceHandler() {
					public ProtocolChain poll() {
						ProtocolChain protocolChain = protocolChains.poll();

						if (protocolChain == null) {
							protocolChain = new DefaultProtocolChain();
							protocolChain.addFilter(new ReadFilter());
							protocolChain.addFilter(filter);
						}

						return protocolChain;
					}
				});
		grizzlyController.addSelectorHandler(tcpHandler);
		grizzlyController.start();
	}

	public void close() {

		this.localAddr = null;

		try {
			grizzlyController.stop();
		} catch (Exception ex) {
			trace(ex);
		}
	}

	public String getServerId() {
		return serverId;
	}

	public InetSocketAddress getLocalAddr() {
		return localAddr;
	}

	public void register(String objectId, RequestHandler handler) {
		handlers.put(objectId, handler);
	}

	public void unregister(String objectId) {
		handlers.remove(serverId);
	}

	private void handleAtomicRequest(String msg) {

		try {
			int pos = msg.indexOf("\n");
			String clientId = msg.substring(0, pos);

			String content = msg.substring(pos + 1);
			pos = content.indexOf("\n");
			String serverObjectId = content.substring(0, pos);
			String msgBody = content.substring(pos + 1);

			RequestHandler handler = handlers.get(serverObjectId);
			if (handler != null) {
				String rsp = handler.handle(msgBody);
				sendResponse(clientId, rsp);
			}
		} catch (Exception ex) {
			trace(ex);
		}
	}

	private void sendResponse(String clientId, String msg) {

		String rspMsg = serverId + "\n" + msg + PAYLOAD_DELIMITER;

		InetSocketAddress clientIdAddr = resolveAddress(clientId);

		ConnectorHandler clientConnector = grizzlyController
				.acquireConnectorHandler(Controller.Protocol.TCP);

		// Initiates new connection or reuses one from cache
		try {

			while (true) {
				try {
					clientConnector
							.connect(clientIdAddr, new BlockingCallbackHandler(
									clientConnector, rspMsg));
					break;
				} catch (AlreadyConnectedException ex) {
					trace("partial failure: already connected");
					// intermittent - expected ??
				}
			}

			// If limit is not reached - connection will be put back to the
			// cache
			try {
				clientConnector.close();
			} finally {
				grizzlyController.releaseConnectorHandler(clientConnector);
			}

		} catch (IOException ex) {
			trace(ex);
		}
	}

	private class MessageDecoder implements ProtocolFilter {

		public boolean execute(Context context) throws IOException {
			final WorkerThread workerThread = ((WorkerThread) Thread
					.currentThread());
			ByteBuffer buffer = workerThread.getByteBuffer();
			buffer.flip();

			if (buffer.hasRemaining()) {
				handleRequest(readBuffer(buffer), context);
			}
			buffer.clear();
			return false;
		}

		public boolean postExecute(Context context) throws IOException {
			return true;
		}

		private void handleRequest(String msg, Context context)
				throws IOException {

			String msgBuffer = (String) context.getAttribute("msgBuffer");
			if (msgBuffer == null) {
				msgBuffer = "";
			}

			msgBuffer = msgBuffer + msg;

			if (msgBuffer.indexOf(PAYLOAD_DELIMITER) != -1) {
				String[] messages = msgBuffer.split(PAYLOAD_DELIMITER);

				for (int i = 0; i < messages.length - 1; i++) {
					handleAtomicRequest(messages[i]);
				}

				String lastMessage = messages[messages.length - 1];
				if (msgBuffer.endsWith(PAYLOAD_DELIMITER)) {
					handleAtomicRequest(lastMessage);
					msgBuffer = "";
				} else {
					msgBuffer = lastMessage;
				}
			}

			context.setAttribute("msgBuffer", msgBuffer);
		}
	}

	private static class BlockingCallbackHandler implements
			CallbackHandler<Context> {

		private ConnectorHandler connector;

		private String msg;

		public BlockingCallbackHandler(ConnectorHandler connector, String msg) {
			this.connector = connector;
			this.msg = msg;
		}

		public void onConnect(IOEvent<Context> ioEvent) {

			SelectionKey key = ioEvent.attachment().getSelectionKey();
			connector.finishConnect(key);

			connector.getController().registerKey(key, SelectionKey.OP_WRITE,
					Controller.Protocol.TCP);
		}

		public void onRead(IOEvent<Context> ioEvent) {
		}

		public void onWrite(IOEvent<Context> ioEvent) {
			try {
				connector.write(ByteBuffer.wrap(msg.getBytes()), true);
			} catch (IOException ex) {
				trace(ex);
			}
		}
	}

}
